import os
from datetime import datetime, timedelta
import subprocess
import  fileinput
import logging
import pdb
import datetime as dt

from airflow import DAG
from airflow.models import Variable

from airflow.operators.bash_operator import BashOperator
from airflow.operators.python_operator import PythonOperator
from airflow.operators.python_operator import BranchPythonOperator
from airflow.operators.subdag_operator import SubDagOperator
from airflow.operators.dummy_operator import DummyOperator

from AGLOW.airflow.operators.LTA_staging import LOFARStagingOperator
from AGLOW.airflow.operators.data_staged import Check_staged

from AGLOW.airflow.subdags.SKSP_calibrator import calibrator_subdag
from AGLOW.airflow.subdags.SKSP_target import target_subdag
from AGLOW.airflow.subdags.SKSP_juelich import juelich_subdag
from AGLOW.airflow.subdags.stage_subdag import test_state_subdag

#Import helper fucntions 
from AGLOW.airflow.utils.AGLOW_MySQL_utils import SurveysDB
from AGLOW.airflow.utils.AGLOW_MySQL_utils import update_field_status_from_taskid
from AGLOW.airflow.utils.AGLOW_MySQL_utils import update_OBSID_status_from_taskid
from AGLOW.airflow.utils.AGLOW_MySQL_utils import get_next_pref
from AGLOW.airflow.utils.AGLOW_MySQL_utils import get_AGLOW_field_properties
from AGLOW.airflow.utils.AGLOW_utils import count_files_uberftp 
from AGLOW.airflow.utils.AGLOW_utils import count_grid_files
from AGLOW.airflow.utils.AGLOW_utils import stage_if_needed
from AGLOW.airflow.utils.AGLOW_utils import get_next_field
from AGLOW.airflow.utils.AGLOW_utils import set_field_status_from_taskid
from AGLOW.airflow.utils.AGLOW_utils import get_srmfile_from_dir
from AGLOW.airflow.utils.AGLOW_utils import count_from_task
from AGLOW.airflow.utils.AGLOW_utils import get_field_location_from_srmlist
from AGLOW.airflow.utils.AGLOW_utils import set_field_status_from_task_return
from AGLOW.airflow.utils.AGLOW_utils import modify_parset_from_fields_task 
from AGLOW.airflow.utils.AGLOW_utils import check_folder_for_files_from_task 

import GRID_LRT


default_args = {
    'owner': 'apmechev',
    'depends_on_past': False,
    'start_date': dt.datetime(2018, 8, 1),
    'email': ['apmechev@gmail.com'],
    'email_on_failure': False, 
    'email_on_retry': False,
    'retries': 0,
    'retry_delay': timedelta(minutes=5),
    'provide_context': True,
    'run_as_user': 'apmechev'
    # 'queue': 'bash_queue',
    # 'pool': 'backfill',
    # 'priority_weight': 10,
    # 'end_date': datetime(2016, 1, 1),
}


dag = DAG('SKSP_Launcher', default_args=default_args, schedule_interval='0 */24 * * *' , catchup=False)

data_directory = "/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/"

if not os.path.isdir(data_directory):
    data_directory = GRID_LRT.__file__.split('__init__.py')[0]+"/data/"

args_dict = { 
            "cal1_parset"   : data_directory+"parsets/Pre-Facet-Calibrator-1.parset",
            "cal2_parset"   : data_directory+"parsets/Pre-Facet-Calibrator-2.parset",
            'pref_cal1_cfg' : data_directory+"config/steps/pref_cal1.cfg",
            'pref_cal2_cfg' : data_directory+"config/steps/pref_cal2.cfg",
            "targ1_parset"  : data_directory+"parsets/Pre-Facet-Target-1.parset",
            "targ2_parset"  : data_directory+"parsets/Pre-Facet-Target-2.parset",
            'pref_targ1_cfg': data_directory+"config/steps/pref_targ1.cfg",
            'pref_targ2_cfg': data_directory+"config/steps/pref_targ2.cfg"
            }


args_dict_juelich = {
                "cal1_parset"   : data_directory+"parsets/Pre-Facet-Calibrator-1.parset",
                "cal2_parset"   : data_directory+"parsets/Pre-Facet-Calibrator-2.parset",
                "targ1_parset"  : data_directory+"parsets/Pre-Facet-Target-1.parset",
                'pref_cal1_cfg' : data_directory+"config/steps/pref_cal1_juelich.cfg",
                'pref_cal2_cfg' : data_directory+"config/steps/pref_cal2.cfg",
                'pref_targ1_cfg': data_directory+"config/steps/pref_targ1.cfg"}

#Resetting variables (They are set by get_srmfiles)

orig_parsets = { 
        "Pre-Facet-Calibrator-1.parset":data_directory+"parsets/Pre-Facet-Calibrator-1.parset", 
        "Pre-Facet-Calibrator-2.parset":data_directory+"parsets/Pre-Facet-Calibrator-2.parset",
        "Pre-Facet-Target-1.parset" : data_directory+'parsets/Pre-Facet-Target-1.parset',
        "Pre-Facet-Target-2.parset" : data_directory+'parsets/Pre-Facet-Target-2_fields.parset'
        }
time_avg = 8
freq_avg = 2


def get_dummy_field(**context):
    return {"field_name":"P252_35"}

def dummy_field_props(_,**context):
    return{
            "baseline_filter": "", "calib_freq_resolution": 16, "targ_time_resolution": 1, "calib_time_resolution": 1, "targ_freq_resolution": 16, "calibrator_nsb": "231", "field_name": "P252_35", "target_nsb": "231", "calib_OBSID": "L654742", "target_OBSID": "L654748"
            }

def select_LTA_location(get_LTA_task,sara_task,juelich_task,**context):
    location = context['task_instance'].xcom_pull(task_ids=get_LTA_task)
    logging.info(location)
    if location == 'sara':
        return sara_task
    if location == 'juelich':
        return juelich_task

params={
    'LRT_config_file': '/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/config/NDPPP_parset.cfg',
    'srmfile':"/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/tests/srm_50_sara.txt",
    'NDPPP_parset':'/home/apmechev/AGLOW/NDPPP.parset'
        }


#test_subdag = SubDagOperator(
#        task_id = 'test_staging_subdag',
#        subdag = test_state_subdag('SKSP_Launcher','test_staging_subdag', default_args, args_dict=params),
#        dag=dag
#        )


get_next_field = PythonOperator(
        task_id = 'get_next_field',
        python_callable=get_next_pref,
        op_args=[],
        dag = dag)

get_field_properties = PythonOperator(
        task_id = 'get_field_properties',
        provide_context=True,
        python_callable=get_AGLOW_field_properties,
        op_args=['get_next_field'],
        dag = dag)


start_field = PythonOperator(
        task_id = 'start_field',
        provide_context=True,
        python_callable=update_OBSID_status_from_taskid,
        op_args=['get_next_field','get_field_properties', 'DI_started'],
        dag=dag)

get_srmfiles = PythonOperator(
        task_id = 'get_srmfiles',
        provide_context = True,
        python_callable = get_srmfile_from_dir,
        op_args = ['/home/timshim/GRID_LRT3/GRID_LRT/SKSP/srmfiles/','get_field_properties'],
        dag = dag)

make_parsets = PythonOperator(
        task_id = 'make_parsets',
        provide_context = True,
        python_callable = modify_parset_from_fields_task,
        op_args = [orig_parsets, 'get_field_properties', time_avg, freq_avg],
        dag = dag
        )

get_LTA_location = PythonOperator(
        task_id = 'get_LTA_location',
        python_callable = get_field_location_from_srmlist,
        op_args = ['get_srmfiles', 'targ_srmfile'],
        dag = dag)

launch_at_lta_location = BranchPythonOperator(
        task_id = 'launch_at_lta_location',
        python_callable = select_LTA_location,
        provide_context = True,
        op_args = ['get_LTA_location','launch_at_sara','launch_at_juelich'],
        dag = dag
        )

    
launch_at_sara = DummyOperator(
        task_id='launch_at_sara',
        dag=dag
        )   

launch_at_juelich = DummyOperator(
       task_id='launch_at_juelich',      
       dag=dag
        )   
        

###########Calibrator branches#
branch_if_cal_exists = BranchPythonOperator(
    task_id = 'branch_if_cal_exists',       
    provide_context = True,                   # Allows to access returned values from other tasks
    python_callable = count_from_task,      
    op_args = ['get_srmfiles', 'cal_srmfile', 'check_calstaged', 'cal_done_already',
        "SKSP",'pref_cal2',1,False],         
    dag = dag)
    

branching_cal = BranchPythonOperator(
    task_id='branch_if_staging_needed',
    provide_context=True,                   # Allows to access returned values from other tasks
    python_callable=stage_if_needed,
    op_args=['check_calstaged','files_staged','stage_cal'],
    dag=dag)
    
files_staged = DummyOperator(
    task_id='files_staged',
    dag=dag
)   
    
calib_done = PythonOperator(
        task_id = 'cal_done',
        provide_context = True,
        python_callable = update_OBSID_status_from_taskid,
        op_args = ['get_next_field','get_field_properties',  'DI_cal_done'],
        dag = dag)

join = DummyOperator(
    task_id='join',
    trigger_rule='one_success',
    dag=dag
)

cal_done_already = DummyOperator(task_id='cal_done_already',
        dag=dag)

join_cal = DummyOperator(task_id='join_cal',
            trigger_rule='one_success',
            dag=dag)


#####################################
#######Calibrator 1 block
#####################################
        
#Stage the files from the srmfile
stage = LOFARStagingOperator( task_id='stage_cal',
            srmfile="SKSP_Prod_Calibrator_srm_file",
            dag=dag)

check_calstaged = Check_staged( task_id='check_calstaged' ,
        srmfile={'name':'get_srmfiles','parent_dag':False},
        srmkey='cal_srmfile',
        dag=dag)

#check_calstaged = Check_staged( task_id='check_calstaged' ,
#            srmfile="SKSP_Prod_Calibrator_srm_file",
#            dag=dag) 
    
            

launch_sara_calibrator = SubDagOperator(
        task_id = 'launch_sara_calibrator',
        subdag = calibrator_subdag('SKSP_Launcher','launch_sara_calibrator', default_args, args_dict=args_dict),
        email_on_failure=True,
        email='apmechev@gmail.com',
        dag=dag
        )

branch_targ_if_staging_needed = BranchPythonOperator(  
    task_id='branch_targ_if_staging_needed',
    provide_context=True,                   # Allows to access returned values from other tasks
    python_callable=stage_if_needed,
    op_args=['check_targstaged','files_staged_targ','stage_targ'],
    dag=dag) 
        
files_staged_targ = DummyOperator(
    task_id='files_staged_targ',
    dag=dag
)   
    
join_targ = DummyOperator(
    task_id='join_targ',
    trigger_rule='one_success',
    dag=dag
)   
     
stage_targ= LOFARStagingOperator( task_id='stage_targ',
        srmfile="SKSP_Prod_Target_srm_file",
        dag=dag)
    
#check_targstaged = Check_staged( task_id='check_targstaged' ,
#        srmfile="SKSP_Prod_Target_srm_file",
#        dag=dag)

check_targstaged = Check_staged( task_id='check_targstaged' ,
        srmfile={'name':'get_srmfiles','parent_dag':False},
        srmkey='targ_srmfile',                  
        dag=dag)



launch_sara_target = SubDagOperator(
        task_id = 'launch_sara_target',
        subdag = target_subdag('SKSP_Launcher','launch_sara_target', default_args, args_dict=args_dict),
        dag=dag
        )

targ_processed = PythonOperator(
        task_id = 'targ_processed',
        provide_context = True,
        python_callable = update_OBSID_status_from_taskid,
        trigger_rule='one_success',
        op_args = ['get_next_field','get_field_properties',  'DI_Processed'],
        dag = dag)

launch_juelich = SubDagOperator(
        task_id = 'launch_juelich',
        subdag = juelich_subdag('SKSP_Launcher','launch_juelich', default_args, args_dict=args_dict_juelich),
        email_on_failure=True,
        email='apmechev@gmail.com',
        dag=dag
        )

#Setting up the dependency graph of the workflow


#Branch if calibrator already processed
get_next_field >> get_field_properties >> start_field >> make_parsets >> get_srmfiles 

get_srmfiles >>  get_LTA_location  

get_LTA_location >>  launch_at_lta_location
launch_at_lta_location >> launch_at_sara
launch_at_lta_location >> launch_at_juelich

launch_at_juelich >> launch_juelich
launch_at_sara  >> branch_if_cal_exists >> cal_done_already >> join_cal >> calib_done

branch_if_cal_exists >> check_calstaged >> branching_cal
branching_cal >> files_staged >> join 
branching_cal >> stage >> join 

join >> launch_sara_calibrator >> join_cal 

launch_at_sara >> check_targstaged >> branch_targ_if_staging_needed 
branch_targ_if_staging_needed >> files_staged_targ >> join_targ
branch_targ_if_staging_needed >>stage_targ >> join_targ

join_targ >> launch_sara_target
calib_done >> launch_sara_target
launch_sara_target >> targ_processed 

launch_juelich >> targ_processed



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from past.builtins import basestring

from datetime import datetime
import logging
from urllib.parse import urlparse
from time import sleep
import re
import sys
import subprocess
import pdb

import airflow
from airflow import hooks, settings
from airflow.exceptions import AirflowException, AirflowSensorTimeout, AirflowSkipException
from airflow.models import BaseOperator, TaskInstance
from airflow.hooks.base_hook import BaseHook
from airflow.hooks.hdfs_hook import HDFSHook
from airflow.utils.state import State
from airflow.operators.sensors import BaseSensorOperator
from airflow.utils.decorators import apply_defaults


class dcacheSensor(BaseSensorOperator):
    """
    Runs a sql statement until a criteria is met. It will keep trying until
    sql returns no row, or if the first cell in (0, '0', '').

    :param conn_id: The connection to run the sensor against
    :type conn_id: string
    :param sql: The sql to run. To pass, it needs to return at least one cell
        that contains a non-zero / empty string value.
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#7c7287'

    @apply_defaults
    def __init__(self, 
            token_task, 
            success_threshold=0.9, 
            poke_interval=300,
            timeout=60*60*24*4, 
            parent_dag=False,
            *args, **kwargs):
        self.token_task = token_task
        self.threshold = success_threshold
        self.parent_dag = parent_dag
        self.glite_status='Waiting'
        super(dcacheSensor, self).__init__(poke_interval=poke_interval,
                timeout=timeout, *args, **kwargs)

    def poke(self, context):
        if not hasattr(self,'dcache_location'):
            self.get_picas_values(context)
        g_proc = subprocess.Popen(['uberftp','-ls', self.dcache_location] ,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        g_result = g_proc.communicate() 
        num_done = self.parse_uberftpls(g_result[0])
        if num_done > self.num_jobs * self.threshold:
            return None
        else:
            logging.info("Only %d jobs done out of %d"%(num_done,self.num_jobs))
            return False

    def parse_uberftpls(self,result):
        num_links = sum(1 for link in str(result).split(b'\n') if len(link) > 1)
        return num_links 

    def get_picas_values(self, context):
        t_task = context['task_instance'].xcom_pull(task_ids=self.token_task)
        self.dcache_location = t_task['output_dir']
        self.num_jobs = t_task['num_jobs']
        if self.num_jobs == 0:
            raise RuntimeError("Zero Jobs expected from  "+str(self.token_task)+" task. ")
        logging.info('Checking files in : ' + self.dcache_location)




from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from past.builtins import basestring

from datetime import datetime
import logging
from urllib.parse import urlparse
from time import sleep
import re
import sys
import subprocess
import pdb

import airflow
from airflow import hooks, settings
from airflow.exceptions import AirflowException, AirflowSensorTimeout, AirflowSkipException
from airflow.models import BaseOperator, TaskInstance
from airflow.hooks.base_hook import BaseHook
from airflow.hooks.hdfs_hook import HDFSHook
from airflow.utils.state import State
from airflow.operators.sensors import BaseSensorOperator
from airflow.utils.decorators import apply_defaults


class gliteSensor(BaseSensorOperator):
    """
    An sensor initialized with the glite-wms job ID. It tracks the status of the job and 
    returns only when all the jobs have exited (finished OK or not)

    :param submit_task: The task which submitted the jobs (should return a glite-wms job ID)
    :type submit_task: string
    :param success_threshold: Currently a dummy
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#7c7287'

    @apply_defaults
    def __init__(self, 
            submit_task, 
            success_threshold=0.9, 
            poke_interval=120,
            timeout=60*60*24*4, 
            *args, **kwargs):
        self.submit_task= submit_task
        self.threshold=success_threshold
        self.glite_status='Waiting'
        super(gliteSensor, self).__init__(poke_interval=poke_interval,
                timeout=timeout, *args, **kwargs)

    def poke(self, context):
        """Function called every (by default 2) minutes. It calls glite-wms-job-status
        on the jobID and exits if all the jobs have finished/crashed. 

        """
        self.jobID=context['task_instance'].xcom_pull(task_ids=self.submit_task)
        if self.jobID==None:
            raise RuntimeError("Could not get the jobID from the "+str(self.submit_task)+" task. ")
        logging.info('Poking glite job: ' + self.jobID)
        g_proc = subprocess.Popen(['glite-wms-job-status', self.jobID] ,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        g_result=g_proc.communicate() 
        self.parse_glite_jobs(g_result[0])
        if not 'Done' in self.job_status:
            if 'Abort' in self.job_status:
                logging.warn("Job aborted from commandline")
                return True 
            return False
        else:
            exit_codes=self.count_successes(g_result[0])
            success_rate=1
            logging.info(str(success_rate)+" of jobs completed ok")
            if (success_rate < self.threshold):
                logging.warn("Less than "+str(self.threshold)+" jobs finished ok!")
            return True

    def parse_glite_jobs(self,jobs): 
        try:
            self.job_status=jobs.split('Current Status:')[1].split()[0]
        except:
            logging.info(jobs)
        logging.debug("Current job status is "+str(self.job_status))
        if self.glite_status== 'Running': 
            self.ui_color='#ef7f23'
        if self.glite_status=='Waiting':
            self.count_successes(jobs)
        if self.glite_status=='Running' and self.job_status=='Waiting':
            self.glite_status='Completed'


    def count_successes(self,jobs):
        """Counts the number of Completed jobs in the results of the glite-wms-job-status
        output. Returns all the job statuses and sets self.job_status if it's Done
        
        :param jobs: A string containing the full output of glite-wms-job-status
        :type jobs: str
        """
        exit_codes=[]
        jobs_list=[]
        for j in jobs.split('=========================================================================='):
            jobs_list.append(j)
        statuses=[]
        for j in jobs_list:
            if "Current Status:" in j:
                statuses.append(j.split("Current Status:")[1].split('\n')[0])
        numdone=0
        for i in statuses:
            if 'Done' in i or 'Cancelled' in i or 'Aborted' in i  :
                numdone+=1
        if 'Done' in statuses[0]: 
            self.job_status = 'Done'
        if numdone == len(jobs):
            self.job_status='Done'
        if self.job_status == 'Waiting':
            for i in statuses:
                if 'Scheduled' in i or 'Running' in i:
                    self.job_status = 'Waiting'
                    return  statuses[1:]
                self.job_status = "Done"
        logging.info("Num_jobs_done "+str(numdone))
        return statuses[1:]
        



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from past.builtins import basestring

from datetime import datetime
import logging
from urllib.parse import urlparse
from time import sleep
import re
import os
import sys
import subprocess
import pdb

import airflow
from airflow import hooks, settings
from airflow.exceptions import AirflowException, AirflowSensorTimeout, AirflowSkipException
from airflow.models import BaseOperator, TaskInstance
from airflow.hooks.base_hook import BaseHook
from airflow.hooks.hdfs_hook import HDFSHook
from airflow.utils.state import State
from airflow.operators.sensors import BaseSensorOperator
from airflow.utils.decorators import apply_defaults
from airflow.models import Variable
from AGLOW.airflow.utils.AGLOW_utils import get_task_instance                                                               


from GRID_LRT.Staging import state_all

class Check_staged(BaseOperator):
    """
    TODO: srmls -l on the base directory and parse locality instead of looping!


    :param conn_id: The connection to run the sensor against
    :type conn_id: string
    :param sql: The sql to run. To pass, it needs to return at least one cell
        that contains a non-zero / empty string value.
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#7c7287'

    @apply_defaults
    def __init__(self, 
            srmfile=None, 
            srmkey=None,
            success_threshold=0.9,  
            *args, **kwargs):
        if srmfile:
            self.srmfile=srmfile
        else:
            raise ValueError("srmfile not defined") 
        self.threshold=success_threshold
        self.srmkey = srmkey
        super(Check_staged, self).__init__(*args, **kwargs)

    def execute(self, context):
        if isinstance(self.srmfile,dict):
            task_name = self.srmfile['name']
            task_parent_dag = self.srmfile['parent_dag']
            sbx_xcom = get_task_instance(context, task_name, task_parent_dag)
            self.srmfile = sbx_xcom[self.srmkey]
        elif "/" not in self.srmfile:
            self.srmfile= Variable.get(self.srmfile)
        logging.info(self.srmfile)
        staging_statuses=state_all.main(self.srmfile, verbose=False)
        logging.info(staging_statuses)
        if state_all.percent_staged(staging_statuses) > self.threshold:
            return {'staged':True,'srmfile':str(self.srmfile)}
        return {'staged':False,'srmfile':str(self.srmfile)}


    def determine_srm_root_dir(self,srmfile):
        """
        Uses os and path to determine the root dir of [THE FIRST] srm link in the file
        """
        srmlink=""
        for line in open(srmfile).readlines():
            if line:
                srmlink=line
                break
        srmdir=os.path.dirname(srmlink.split()[0])
        if srmdir=="":
            raise ValueError("SRM directory cannot be empty")
        logging.info("Found SRM directory: "+str(srmdir))
        return srmdir

    def check_srm_status(self,srmdir):
        """Checks the status of all srm links in the folder given. 
        Returns a list of localities
        """
        g_proc = subprocess.Popen(['srmls','-l', srmdir] ,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out=g_proc.communicate()
        if out[1]!='':
            raise RuntimeError('srmls failed')
        localities=[i.split(':')[1] 
                    for i in out[0].split('\n') 
                    if 'locality' in i] 
        return localities[1:]  # The first element is the locality of the folder
                    
    def count_successes(self,status_list):
        suc=sum(st=='ONLINE_AND_NEARLINE' for st in status_list)
        fail=sum(st=='NEARLINE' for st in status_list)
        logging.info(str(suc/float(suc+fail))+ " of the files are staged")
        return suc/float(suc+fail)

    def check_gfal_status(self,srmlist=None):
        """
        Uses gfal instead of srmls to find out number of staged files
        """
        pass



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from builtins import bytes
import os
import os.path
import signal
from time import sleep
import logging
from subprocess import Popen, STDOUT, PIPE
from tempfile import gettempdir, NamedTemporaryFile


from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.utils.file import TemporaryDirectory
from airflow.utils.state import State
#import progressbar
#logging.info(progressbar.__file__)
import GRID_LRT.sandbox as Sandbox

import pdb

class LRTSandboxOperator(BaseOperator):
    """
    Operator that uploads a LOFAR 'sandbox' to GRID storage

    :param sbx_config: path to the configuration file the describes how to build the sandbox.
    :type sbx_config: str
    :param tok_config: path to the configuration file for the tokens. Optional
    :type tok_config: str
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#42f4a4'

    @apply_defaults
    def __init__(
            self,
            sbx_config,
            tok_config=None,
            output_encoding='utf-8',
            *args, **kwargs):

        super(LRTSandboxOperator, self).__init__(*args, **kwargs)
        self.sbx_config = sbx_config
        self.tok_config = tok_config  
        self.output_encoding = output_encoding
        self.state=State.QUEUED

    def build_sandbox(self):
        """Do not build Sandbox on initialization but on execution"""
        self.SBX=Sandbox.Sandbox(cfgfile=self.sbx_config)


    def execute(self, context):
        """
        Launches the task. Leverages the sandbox class to build
        a sandbox using a configuration file. Sandbox is uploaded to 
        the respective storage. Xcom returned is the location of the
        Sandbox. The sandbox is cleaned up locally.
        """
        self.build_sandbox()
        SBXlocs=[]
        self.SBX.build_sandbox(self.sbx_config)        
        self.SBX.upload_sandbox()
        self.SBX.cleanup()
        return {"SBX_location":self.SBX.sbxloc}


    def get_SBXloc(self):
        return self.SBX.SBXloc

    def on_kill(self):
        """Cleans up the sandbox before killing process"""
        logging.warn('Sending SIGTERM signal to staging group')
        self.state=State.SHUTDOWN
        self.SBX.cleanup()
        os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM)



import os
import subprocess
import fileinput
import logging
from datetime import datetime, timedelta


from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from airflow.operators.python_operator import PythonOperator
from airflow.operators.python_operator import BranchPythonOperator
from airflow.operators.dummy_operator import DummyOperator
from airflow.models import Variable

#Import AGLOW operators
from AGLOW.airflow.operators.LTA_staging import LOFARStagingOperator
from AGLOW.airflow.operators.LRT_Sandbox import LRTSandboxOperator
from AGLOW.airflow.operators.LRT_token import TokenCreator,TokenUploader,ModifyTokenStatus
from AGLOW.airflow.operators.LRT_submit import LRTSubmit 
from AGLOW.airflow.operators.data_staged import Check_staged
from AGLOW.airflow.operators.LRT_storage_to_srm import Storage_to_Srmlist 
from AGLOW.airflow.sensors.glite_wms_sensor import gliteSensor

#Import helper fucntions 
from AGLOW.airflow.utils.AGLOW_utils import get_next_field
from AGLOW.airflow.utils.AGLOW_utils import count_files_uberftp 
from AGLOW.airflow.utils.AGLOW_utils import count_grid_files
from AGLOW.airflow.utils.AGLOW_utils import stage_if_needed
from AGLOW.airflow.utils.AGLOW_utils import get_next_field
from AGLOW.airflow.utils.AGLOW_utils import set_field_status_from_taskid
from AGLOW.airflow.utils.AGLOW_utils import get_srmfile_from_dir
from AGLOW.airflow.utils.AGLOW_utils import count_from_task
from AGLOW.airflow.utils.AGLOW_utils import get_field_location_from_srmlist
from AGLOW.airflow.utils.AGLOW_utils import set_field_status_from_task_return
from AGLOW.airflow.utils.AGLOW_utils import modify_parset_from_fields_task 
from AGLOW.airflow.utils.AGLOW_utils import check_folder_for_files_from_task 

#Import GRID_LRT helpers
from GRID_LRT.Staging import state_all
from GRID_LRT.Staging.srmlist import srmlist
from GRID_LRT import Token
from GRID_LRT.get_picas_credentials import picas_cred




#TODO: This fails!

def archive_tokens_from_task(token_task, delete=False, **context):
    """ Determines whic tokens to archive and saves them. delete if necessary
    """
    task_dict = context['ti'].xcom_pull(token_task)
    t_type = task_dict['token_type']
    archive_location = task_dict['output_dir']
    archive_all_tokens(t_type, archive_location, delete=delete)


def archive_all_tokens(token_type, archive_location, delete=False):
    pc = picas_cred()
    th = Token.Token_Handler(t_type=token_type, uname=pc.user, pwd=pc.password, dbn=pc.database)
    token_archive = th.archive_tokens(delete_on_save=delete, compress=True)
    logging.info("Archived tokens from " + token_type + " and made an archive: " + token_archive)
    logging.info(token_archive + " size is " + str(os.stat(token_archive).st_size))
    subprocess.call(['globus-url-copy '+token_archive+" "+archive_location+"/"+token_archive.split('/')[-1]],shell=True)
    logging.info("Resulting archive is at "+archive_location+"/"+token_archive.split('/')[-1])

    

def target_subdag(parent_dag_name, subdagname,dag_args, args_dict=None):
    field_name = 'fields_'
    #Resetting variables (They are set by get_srmfiles)
    
    dag = DAG(dag_id=parent_dag_name+'.'+subdagname, default_args=dag_args, schedule_interval='@once' , catchup=False)
    if not args_dict:
        args_dict = { 
            "targ1_parset":"/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/parsets/Pre-Facet-Target-1.parset", 
            "targ2_parset":"/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/parsets/Pre-Facet-Target-2.parset", 
            'pref_targ1_cfg':'/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/config/steps/pref_targ1.cfg',
            'pref_targ2_cfg':'/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/config/steps/pref_targ2.cfg'
            }

    #Create a sandbox for the job
    sandbox_targ = LRTSandboxOperator(task_id='sbx',
            sbx_config=args_dict['pref_targ1_cfg'],
            dag=dag)
    
    #Create the tokens and populate the srm.txt 
    tokens_targ = TokenCreator(task_id='token_targ',
            sbx_task={'name':'sbx','parent_dag':False},
            staging_task ={'name':'check_targstaged','parent_dag':True},
            srms_task={'name':'get_srmfiles','parent_dag':True},
            token_type=field_name,
            tok_config =args_dict['pref_targ1_cfg'],
            files_per_token=1,
            dag=dag)
    
    #Upload the parset to all the tokens
    parset_targ = TokenUploader(task_id='targ_parset', 
            token_task='token_targ',
            upload_file=args_dict['targ1_parset'],
            parset_task = 'make_parsets',
            parent_dag = True,
            dag=dag)
    
    
    #Submit tokens to the GRID
    submit_targ = LRTSubmit(task_id='submit',
            token_task='token_targ',
            parameter_step=1,
            NCPU=2,
            dag=dag)
    
    #Wait for all jobs to finish
    wait_for_run_targ = gliteSensor(task_id='running',
            submit_task='submit',
            success_threshold=0.95,
            poke_interval=120,
            dag=dag)
        
    targ1_files_done = PythonOperator(
            task_id = 'targ1_files_done',
            python_callable = check_folder_for_files_from_task,
            op_args = ['token_targ','output_dir', 230],
            dag=dag
            )
    #####################################
    #######Target 2 block
    #####################################
    
    sandbox_targ2 = LRTSandboxOperator(task_id='sbx_targ2', 
            sbx_config=args_dict['pref_targ2_cfg'],
            dag=dag)
   
    targ2_srmlist_from_storage = Storage_to_Srmlist(task_id='srmlist_from_targ1',
              token_task='token_targ',
              dag=dag)

    tokens_targ2 = TokenCreator( task_id='token_targ2',
            staging_task={'name':'srmlist_from_targ1','parent_dag':False},
            sbx_task={'name':'sbx_targ2','parent_dag':False},
            token_type=field_name,                                                                           
            files_per_token=10,
            subband_prefix='AB',
            tok_config=args_dict['pref_targ2_cfg'],
            dag=dag)
    
    parset_targ2 = TokenUploader( task_id='targ_parset2',
            token_task='token_targ2',
            upload_file=args_dict['targ2_parset'],
            parset_task = 'make_parsets',
            parent_dag = True,
            dag=dag)
    
    submit_targ2 = LRTSubmit( task_id='submit_targ2',
            token_task='token_targ2',
            parameter_step=1, 
            NCPU=5,
            dag=dag)
    
    wait_for_run_targ2 = gliteSensor( task_id='running_targ2',
            submit_task='submit_targ2',
            success_threshold=0.9,
            poke_interval=120,
            dag=dag)

    all_files_done = PythonOperator(
        task_id = 'all_files_done',
        python_callable = check_folder_for_files_from_task,
        op_args = ['token_targ2','output_dir', None],
        dag=dag
        )

    archive_targ2 = PythonOperator(
            task_id = 'archive_targ2',
            provide_context=True, 
            python_callable=archive_tokens_from_task,
            op_args=['token_targ2'],
            dag = dag)
     

    ## Setting up the dependency graph of the workflow
    
    
    #Branch if targibrator already processed

    sandbox_targ >> tokens_targ >> parset_targ >>  submit_targ >> wait_for_run_targ
    wait_for_run_targ >> targ1_files_done >> targ2_srmlist_from_storage >> sandbox_targ2 >> tokens_targ2 >> parset_targ2 >> submit_targ2 >> wait_for_run_targ2
    
    wait_for_run_targ2 >> all_files_done >> archive_targ2 
    return dag



from airflow import DAG                                                                                                     
from AGLOW.airflow.operators.LTA_staging import LOFARStagingOperator
from AGLOW.airflow.operators.data_staged import Check_staged                                                                              



def test_state_subdag(parent_dag_name, subdagname, dag_args, args_dict=None):
    dag = DAG(dag_id=parent_dag_name+'.'+subdagname, default_args=dag_args, schedule_interval='@once' , catchup=True)

    if not args_dict:
        args_dict = {'srmfile':'/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/tests/srm_50_sara.txt'}

    check_calstaged = Check_staged( task_id='check_calstaged',srmfile="SKSP_Prod_Target_srm_file",dag=dag)


    stage = LOFARStagingOperator( task_id='stage_cal',
           srmfile=args_dict['srmfile'],
           dag=dag)
    check_calstaged >> stage
    return dag



import sshtunnel
import logging
import socket as sock
#import MySQLdb as msqldb
#import MySQLdb.cursors as mdbcursors
import os
import datetime
from collections import namedtuple
from time import sleep
#sshtunnel.DEFAULT_LOGLEVEL = logging.DEBUG

LocationTuple = namedtuple('LocationTuple',['location','count'])

def count_runs_at_locations(query_results):
    locations = {'Sara':0,'Juelich':0, 'Poznan':0}
    for i in query_results:
        if i['location'] in locations.keys():
                locations[i['location']]+=1
    sorted_locations = sorted(locations.items(), key=lambda kv: kv[1])
    results = []
    for location in sorted_locations:
        if not skip_location(location[0]):
            results.append(LocationTuple(*location))
    return results

def get_next_at_location(location="Juelich"):
    sdb = SurveysDB(readonly=True)
    sdb.cur.execute('select observations.id,observations.field,fields.priority,observations.status,observations.priority,observations.location from observations left join fields on (observations.field=fields.id) where observations.status="Observed" and observations.location = "%s" order by observations.priority desc ,fields.priority desc limit 2'%(location))
    results = sdb.cur.fetchone()
    sdb.close()
    return results

def get_running_observations():
    sdb=SurveysDB(readonly=True)
    sdb.cur.execute('select location,status from observations where observations.status!="Observed" and observations.status!="DI_Processed" and observations.status!="Preprocessed";')
    results=sdb.cur.fetchall()
    return results


def get_next_obsrvation():
    # return the name of the top-priority field with appropriate status
    sdb=SurveysDB(readonly=True)
#    sdb.cur.execute('select fields.id as id,sum(integration) as s,count(observations.id) as c,fields.priority from fields left join observations on (observations.field=fields.id) where fields.status="Not started" and observations.status="DI_processed" group by fields.id having s>7 order by fields.priority desc,ra')
    sdb.cur.execute('select observations.id,observations.field,fields.priority,observations.priority from observations left join fields on (observations.field=fields.id) where observations.status="Observed" order by fields.priority desc,observations.priority desc')
    results = sdb.cur.fetchone()
    sdb.close()
    return results[0]['field']

def skip_location(location='Juelich'):
    """
    Decides if we should skip runs at location if all OBS have Priority <0
    """
    sdb=SurveysDB(readonly=True)
    sdb.cur.execute('select max(priority) from observations where location = \"%s\" '%location)
    results = sdb.cur.fetchone()
    sleep(1)
    sdb.close()
    print("Maximum priority at location %s is %f " % (location, results['max(priority)']))
    if results['max(priority)']== -1:
        return True
    return False

def get_next_pref(**kwargs):
    # return the name of the top-priority field balancing all locations
    running = get_running_observations()
    running_locs = count_runs_at_locations(running)
    next_run_location = running_locs[0].location
    results = get_next_at_location(location=next_run_location)
    return {'field_name':results['field'], 
            'sanitized_field_name':results['field'].replace('+','_')}



def get_AGLOW_field_properties(field_step, **context):
    name = context['ti'].xcom_pull(field_step)['field_name']
    props = get_one_observation(name)
    return {'target_OBSID':'L' + str(int(props['id'])),
            'targ_freq_resolution':int(props['nchan']),
            'targ_time_resolution':int(props['dt']),
            'calib_OBSID':'L'+str(int(props['calibrator_id'])),
            'calib_freq_resolution':int(props['calibrator_nchan']),
            'calib_time_resolution':int(props['calibrator_dt']),
            'field_name':str(props['field']),
            'baseline_filter':str(props['bad_baselines']),
            'calibrator_nsb':str(props['calibrator_nsb']),
            'target_nsb':str(props['nsb'])}

def get_one_observation(field_name):
    sdb=SurveysDB()
    idd=sdb.get_observation(field_name)
    sdb.close()
    return idd

def update_status(field_name, obsid,status,table,time=None):
    # utility function to just update the status of an observation
    # name can be None (work it out from cwd), or string (field name)

    sdb=SurveysDB()
    if table=='fields':
        idd=sdb.get_field(id)
        idd['status']=status
        tag_field(sdb,idd)
        sdb.set_field(idd)
    elif table == 'observations':
        idd=sdb.get_observation(field_name, obsid)
        idd['status']=status
        sdb.set_observation(idd)
    sdb.close()

def update_field_status_from_taskid(task_id, status,  **context):
    field_data=context['ti'].xcom_pull(task_id)
    field_name = field_data['field_name']
    update_status(field_name,status,'fields',time=datetime.datetime.now())

def update_OBSID_status_from_taskid(field_task_id,OBSID_task_id, status, **context):
    field_data = context['ti'].xcom_pull(field_task_id)
    field_name = field_data['field_name']
    obsid_data = context['ti'].xcom_pull(OBSID_task_id)
    obsid = obsid_data['target_OBSID'][1:]
    update_status(field_name, obsid,status,'observations',time=datetime.datetime.now())



def tag_field(sdb,idd):
    # Add location and user tags
    idd['clustername']=get_cluster()
    idd['location']=os.getcwd()
    idd['username']=get_user()
    idd['nodename']=sdb.hostname
    
def get_id():
    dir=os.getcwd()
    dname=dir.split('/')[-1]
    return dname

def get_user():
    return os.getenv('USER')

def get_cluster():
    cluster=os.getenv('DDF_PIPELINE_CLUSTER')
    if cluster:
        return cluster
    else:
        return 'Unknown'

def use_database():
    return 'DDF_PIPELINE_DATABASE' in os.environ

class SurveysDB(object):
    ''' Provides low-level and high-level interfaces to the surveys database '''

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def __init__(self,localport=33306,readonly=False):
        import MySQLdb as msqldb
        import MySQLdb.cursors as mdbcursors

        # get the config file -- this must exist
        home=os.getenv("HOME")
        cfg=open(home+'/.surveys').readlines()
        self.password=cfg[0].rstrip()
        try:
            self.ssh_user=cfg[1].rstrip()
        except:
            self.ssh_user=None
        
        # read only use
        self.readonly=readonly

        # set up an ssh tunnel if not running locally
        self.usetunnel=True
        self.hostname=sock.gethostname()
        if self.hostname=='lofar-server':
            self.con = msqldb.connect('127.0.0.1', 'survey_user', self.password, 'surveys')
        else:
            try:
                dummy=sock.gethostbyname('lofar-server.data')
            except:
                self.usetunnel=True

            if self.usetunnel:
                self.tunnel=sshtunnel.SSHTunnelForwarder('lofar.herts.ac.uk',
                                                         ssh_username=self.ssh_user,
                                                         ssh_pkey=home+'/.ssh/id_rsa_martin',
                                                         remote_bind_address=('127.0.0.1',3306),
                                                         local_bind_address=('127.0.0.1',localport))
                self.tunnel.start()
                self.con = msqldb.connect('127.0.0.1', 'survey_user', self.password, 'surveys', port=localport)
            else:
                self.con = msqldb.connect('lofar-server.data', 'survey_user', self.password, 'surveys')
        self.cur = self.con.cursor(cursorclass=mdbcursors.DictCursor)
        if not self.readonly:
            self.cur.execute('lock table fields write, observations write')
        self.closed=False

    def close(self):
        if not self.closed:
            if not self.readonly:
                self.cur.execute('unlock tables')
            self.con.close()
            if self.usetunnel:
                self.tunnel.stop()
            self.closed=True # prevent del from trying again
    
    def __del__(self):
        self.close()
        
    def get_field(self,id):
        self.cur.execute('select * from fields where id=%s',(id,))
        result=self.cur.fetchall()
        if len(result)==0:
            return None
        else:
            return result[0]

    def set_field(self,sd):
        assert not self.readonly
        id=sd['id'];
        for k in sd:
            if k=='id':
                continue
            if sd[k] is not None:
                self.cur.execute('update fields set '+k+'=%s where id=%s',(sd[k],id))

    def create_field(self,id):
        self.cur.execute('insert into fields(id) values (%s)',(id,))
        return self.get_field(id)

    def get_observation(self,field_id, obsid=None):
        if not obsid:
            self.cur.execute('select * from observations where field=%s',(field_id,))
        else:
            self.cur.execute('select * from observations where field=%s and id=%s',(field_id,obsid))
        result=self.cur.fetchall()
        if len(result)==0:
            return None
        else:
            return result[0]

    def set_observation(self,sd):
        assert not self.readonly
        id=sd['id'];
        for k in sd:
            if k=='id':
                continue
            if sd[k] is not None:
                self.cur.execute('update observations set '+k+'=%s where id=%s',(sd[k],id))

    def create_observation(self,id):
        self.cur.execute('insert into observations(id) values (%s)',(id,))
        return self.get_field(id)

def force_close_tunnel(localport=33306):
    tunnel=sshtunnel.SSHTunnelForwarder('lofar.herts.ac.uk',ssh_username='apmechev',ssh_pkey=os.getenv("HOME")+'/.ssh/id_rsa_martin',remote_bind_address=('127.0.0.1',3306),local_bind_address=('127.0.0.1',localport))
    tunnel.start()
    tunnel.stop()


if __name__=='__main__':
    sdb=SurveysDB()
    result=sdb.get_field('P35Hetdex10')
    #result['location']='Never Never Land'
    #sdb.set_id(result)
    print(result)
    sdb.close()




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from builtins import bytes
import os
import os.path
import signal
from time import sleep
import logging
from subprocess import Popen, STDOUT, PIPE
from tempfile import gettempdir, NamedTemporaryFile


from airflow.exceptions import AirflowException
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from airflow.utils.file import TemporaryDirectory
from airflow.utils.state import State
from GRID_LRT.Staging import stager_access, stage_all_LTA
from GRID_LRT.Staging import srmlist
from AGLOW.airflow.utils.AGLOW_utils import get_var_from_task_decorator
from airflow.models import Variable

try:
    import xmlrpclib
    from xmlrpclib import ResponseError
    from xmlrpclib import Fault as xmlrpcFault
except ImportError:
    import xmlrpc as xmlrpclib
    from xmlrpc.client import ResponseError
    from xmlrpc.client import Fault as xmlrpcFault


class LOFARStagingOperator(BaseOperator):
    """
    Execute a stage command using the LOFAR API. Input is either file with srms
        a list of srms or both.

    :param srmfile: the name of the file holding a list of srm files to stage
    :type srmfile: string
    :param srms: a list of the srms that need to be staged
    :type srms: list
    :param stageID: In case staging was already done
    :type stageID: string
    :type output_encoding: output encoding of bash command
    """
    template_fields = ()
    template_ext = ()
    ui_color = '#f99c2b'

    @apply_defaults
    def __init__(
            self,
            srmfile="",
            srms=None,
            stageID=None,
            output_encoding='utf-8',
            *args, **kwargs):

        super(LOFARStagingOperator, self).__init__(*args, **kwargs)
        self.srmfile = srmfile
        self.srms = srms
        self.output_encoding = output_encoding
        self.state=State.QUEUED
        self.pass_threshold=0.95 #At least 95% need to be staged to continue.
                                 #This corresponds to 10/244 lost files

    def execute(self, context):
        """
        Executes the staging command from the list of srms requested.
        """
        if not os.path.isfile(self.srmfile) and not hasattr(self.srms, '__iter__'):
            self.srmfile=Variable.get(self.srmfile)
            if not os.path.isfile(self.srmfile):
                self.status=State.UPSTREAM_FAILED
                raise AirflowException("Input srmfile doesn't exist and srm list not a list")
        self.progress={'Percent done' : 0}

        surl_list=srmlist.srmlist() #holds all the srms (both from file and list argument )
        self.surl_list=self.build_srm_list(surl_list)
        try:
            self.stage_ID=stager_access.stage(list(self.surl_list))
        except xmlrpcFault : 
            sleep(60)
            self.stage_ID=stager_access.stage(list(self.surl_list))
        logging.info("Successfully sent staging command for " + 
                       stager_access.get_progress()[str(self.stage_ID)]['File count']
                     + " files.")
        logging.info("StageID= " + str(self.stage_ID))

        self.state=State.RUNNING
        sleep(120)
        try:
            self.progress=stager_access.get_progress()[str(self.stage_ID)]
        except:
            pass
        self.started=False
        f=NamedTemporaryFile(delete=False)
        for i in surl_list:
            f.write(i)
        f.close()
        if not f:
            f.name=""
        while self.still_running():  
            sleep(120)
        if self.state==State.SUCCESS:
            return {'srmfile' : str(f.name)}
        self.state==State.FAILED
        return {'srmfile' : str(f.name)}

    def build_srm_list(self,surl_list):
        """Allows the user to input a list of links dynamically 
            or to just give a srmfile or both
        """
        if hasattr(self.srms, '__iter__'):
            for srm in self.srms:
                surl_list.append(srm)
        file_surls=stage_all_LTA.return_srmlist(self.srmfile)

        for srm in file_surls:
            surl_list.append(srm)
        logging.debug(str(len(surl_list)) + " files up for staging. First srm is " +
                                surl_list[0] + " and last one is " + surl_list[-1])
        logging.debug("The OBSID is "+surl_list.OBSID)
        return surl_list


    def not_yet_started(self):
        """
        Checks the status of the staging
        """
        self.staging_status=stager_access.get_status(int(self.stage_ID))
        if self.staging_status == 'new':
            logging.debug("Stage status is 'new'")
            return True

        if self.staging_status == 'scheduled':
            logging.debug("Stage status is 'scheduled'")
            return True
        self.started=True
        return False

    def still_running(self):
        '''Uses the LTA interface to determine the current status
            of the job. it exits true if it's still running (or queued)
            and exits false if it's crashed or completed. the self.status
            is updated accordingly
        '''
        self.staging_status=stager_access.get_status(int(self.stage_ID))
        if self.staging_status =='success':
            self.success()
            return False
        try:
            self.progress=stager_access.get_progress()[str(self.stage_ID)]
        except (KeyError, ResponseError): #After staging is done, the request gets deleted....
            logging.warning("Staging request expired too quickly. This happens if the" +
                "files are already staged (GOOD) or if something goes wrong (BAD). ")
            self.success()
            return False

        if (not self.started) and self.not_yet_started():
            return True # function doesn't evaluate if already started!

        logging.debug(self.progress['Percent done'] + " percent staged.")
        if self.progress['Flagged abort'] == 'true':
            logging.warn('Files flagged for abort by LTA')
            self.status=State.UP_FOR_RETRY
            return False

        if self.staging_status == 'in progress':
            if float(self.progress['Percent done'])/100. < self.pass_threshold:
                logging.debug("Stage status is 'in progress'")
                return True
            else:
                self.success()
                return False
        if self.progress['Status'] == 'on hold':
            return True

    def success(self):
        self.status=State.SUCCESS
        logging.info("Successfully staged the files.")
    
    def return_OBSID(self):
        """
        Will return the OBSID as recorded by the srmlist class
        """
        return self.surl_list.OBSID

    def on_kill(self):
        logging.warn('Sending SIGTERM signal to staging group')
        self.state=State.SHUTDOWN
        os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM) 



#!/usr/bin/env python

#from distutils.core import setup
from setuptools import setup
from setuptools.command.install import install
from AGLOW.version import __version__
import AGLOW
import os


PATCH_LOC='AGLOW/patches'

class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        # PUT YOUR POST-INSTALL SCRIPT HERE or CALL A FUNCTION
        dirs =  get_airflow_directories()
        install.run(self)

class PreInstallCommand(install):
    os.environ['AIRFLOW_GPL_UNIDECODE']='yes'


setup(name='AGLOW',
      packages=['AGLOW','AGLOW/airflow','AGLOW/airflow/dags','AGLOW/airflow/operators','AGLOW/airflow/sensors','AGLOW/airflow/utils', 'AGLOW/airflow/subdags', 'AGLOW/airflow/postgres'],
      version=__version__,
      setup_requires=[
          ],
      tests_require=[
          'pytest', 
          ],
      install_requires=[
          'python-daemon==2.1.2',
          'sshtunnel',
          'snakebite',
          'tzlocal==1.5.1',   
          'apache-airflow',
          'GRID_LRT>=0.4.3.1',
          'six'
          ],
      include_package_data=True,
      data_files=[("logo",["AGLOW/AGLOW_logo.png"]),
                  ("parches",[PATCH_LOC + i for i in os.listdir(PATCH_LOC)])
                 ],
      description='AGLOW: Automated Grid-enabled LOFAR Workflows',
      long_description="Distributing, Automating and Accelerating LOFAR data processing using an industry standard workflow orchestration software on a High Throughput cluster at the Dutch Grid location.",
      author='Alexandar Mechev',
      author_email='LOFAR@apmechev.com',
      url='https://www.github.com/apmechev/AGLOW/',
      download_url = 'https://github.com/apmechev/AGLOW/archive/v0.0.9x.tar.gz',
      keywords = ['surfsara', 'distributed-computing', 'LOFAR', 'workflow','airflow'],
      classifiers = ['Development Status :: 2 - Pre-Alpha',
      'Intended Audience :: Developers',
      'Intended Audience :: Science/Research',
      "License :: OSI Approved :: GNU General Public License v3 (GPLv3)" ,
      "Natural Language :: English",
      "Topic :: Scientific/Engineering :: Astronomy",
      "Topic :: System :: Distributed Computing",
#      'Programming Language :: Python :: 2',
      'Programming Language :: Python :: 2.6',
      'Programming Language :: Python :: 2.7',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.2',
      'Programming Language :: Python :: 3.3',
      'Programming Language :: Python :: 3.4'] 
     )

def get_airflow_directories():
    import airflow
    airflow_file = airflow.__file__
    airflow_dirs = {}
    airflow_dirs['airflow_base_dir'] = airflow_file.split('__init__')[0]
    base_dir = airflow_dirs['airflow_base_dir']
    airflow_dirs['airflow_contrib_dir'] = base_dir + 'contrib/'
    airflow_dirs['airflow_operators_dir'] = base_dir + 'contrib/operators/'
    airflow_dirs['airflow_sensors_dir'] = base_dir + 'contrib/sensors/'       
    airflow_dirs['airflow_utils_dir'] = base_dir + 'utils/'
    for key in airflow_dirs:
        if not os.path.exists(airflow_dirs[key]):
            raise RuntimeError("Directory %s doesn't exist" % airflow_dirs[key])
    return airflow_dirs



from datetime import datetime, timedelta
import subprocess
import fileinput
import logging

from airflow import DAG                                                                                                                     
from airflow.operators.bash_operator import BashOperator
from airflow.operators.python_operator import PythonOperator
from airflow.operators.python_operator import BranchPythonOperator
from airflow.operators.dummy_operator import DummyOperator
from airflow.models import Variable


from AGLOW.airflow.operators.LTA_staging import LOFARStagingOperator
from AGLOW.airflow.operators.LRT_Sandbox import LRTSandboxOperator
from AGLOW.airflow.operators.LRT_token import TokenCreator,TokenUploader,ModifyTokenStatus
from AGLOW.airflow.operators.LRT_submit import LRTSubmit
from AGLOW.airflow.operators.data_staged import Check_staged
from AGLOW.airflow.sensors.dcache_sensor import dcacheSensor
from AGLOW.airflow.operators.LRT_storage_to_srm import Storage_to_Srmlist

from AGLOW.airflow.utils.AGLOW_utils import get_next_field
from AGLOW.airflow.utils.AGLOW_utils import count_files_uberftp
from AGLOW.airflow.utils.AGLOW_utils import count_grid_files
from AGLOW.airflow.utils.AGLOW_utils import stage_if_needed
from AGLOW.airflow.utils.AGLOW_utils import get_next_field
from AGLOW.airflow.utils.AGLOW_utils import set_field_status_from_taskid
from AGLOW.airflow.utils.AGLOW_utils import get_srmfile_from_dir
from AGLOW.airflow.utils.AGLOW_utils import count_from_task
from AGLOW.airflow.utils.AGLOW_utils import get_field_location_from_srmlist
from AGLOW.airflow.utils.AGLOW_utils import set_field_status_from_task_return
from AGLOW.airflow.utils.AGLOW_utils import modify_parset_from_fields_task

from GRID_LRT.Staging.srmlist import srmlist

def archive_tokens_from_task(token_task, delete=False, **context):
    """ Determines whic tokens to archive and saves them. delete if necessary
    """
    task_dict = context['ti'].xcom_pull(token_task)
    t_type = task_dict['token_type']
    archive_location = task_dict['output_dir']
    archive_all_tokens(t_type, archive_location, delete=delete)


def archive_all_tokens(token_type, archive_location, delete=False):
    pc = picas_cred()
    th = Token.Token_Handler(t_type=token_type, uname=pc.user, pwd=pc.password, dbn=pc.database)
    token_archive = th.archive_tokens(delete_on_save=delete, compress=True)
    logging.info("Archived tokens from " + token_type + " and made an archive: " + token_archive)
    logging.info(token_archive + " size is " + str(os.stat(token_archive).st_size))
    subprocess.call(['globus-url-copy '+token_archive+" "+archive_location+"/"+token_archive.split('/')[-1]],shell=True)
    logging.info("Resulting archive is at "+archive_location+"/"+token_archive.split('/')[-1])
                                                                                                                                            

def force_staging(srmfile1, **args):
    srmfile=Variable.get(srmfile1)
    return {'staged':False,'srmfile':str(srmfile)}

def juelich_subdag(parent_dag_name, subdagname,dag_args, args_dict=None):
    field_name = 'fields_'

    dag = DAG(dag_id=parent_dag_name+'.'+subdagname, default_args=dag_args, schedule_interval='@once' , catchup=True)

    if not args_dict:
                args_dict = {
                "cal1_parset":"/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/parsets/Pre-Facet-Calibrator-1.parset",
                "cal2_parset":"/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/parsets/Pre-Facet-Calibrator-2.parset",
                "targ1_parset":"/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/parsets/Pre-Facet-Target-1.parset",
                'pref_cal1_cfg':'/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/config/steps/pref_cal1_juelich.cfg',
                'pref_cal2_cfg':'/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/config/steps/pref_cal2.cfg',
                'pref_targ1_cfg':'/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/config/steps/pref_targ1.cfg'}


    branch_if_cal_exists = BranchPythonOperator(
    task_id = 'branch_if_cal_exists',
    provide_context = True,                   # Allows to access returned values from other tasks
    python_callable = count_from_task,
    op_args = ['get_srmfiles', 'cal_srmfile', 'check_calstaged', 'calib_done',
        "SKSP",'pref_cal2',1,True],
    dag = dag)
        


    branching_cal = BranchPythonOperator(
    task_id='branch_if_staging_needed',       
    provide_context=True,                   # Allows to access returned values from other tasks
    python_callable=stage_if_needed,
    op_args=['check_calstaged','files_staged','stage_cal'],
    dag=dag)

    files_staged = DummyOperator(
    task_id='files_staged',
    dag=dag
)   
    
    calib_done = DummyOperator(
    task_id='calib_done',
    dag=dag
)   
    
    join = DummyOperator(
    task_id='join',
    trigger_rule='one_success',
    dag=dag
)   
    

#####################################
#######Calibrator 1 block
#####################################
    
#Stage the files from the srmfile

    stage = LOFARStagingOperator( task_id='stage_cal',
        srmfile={'name':'get_srmfiles','parent_dag':True},
        srmkey='cal_srmfile',
        dag=dag)

    check_calstaged =  Check_staged( task_id='check_calstaged',
            srmfile={'name':'get_srmfiles','parent_dag':True},
            srmkey='cal_srmfile',
        dag=dag)


    sandbox_cal = LRTSandboxOperator(task_id='sbx',
        sbx_config=args_dict['pref_cal1_cfg'],
        dag=dag)
        
#Create the tokens and populate the srm.txt 
    tokens_cal = TokenCreator(task_id='token_cal',
        sbx_task={'name':'sbx','parent_dag':False},
        staging_task ={'name':'check_calstaged', 'parent_dag':False},
        token_type=field_name,
        tok_config=args_dict['pref_cal1_cfg'],
        pc_database = 'sksp2juelich',
        fields_task = {'name':'get_next_field','parent_dag':True},
        files_per_token=1,
        dag=dag)
        
#Upload the parset to all the tokens
    parset_cal = TokenUploader(task_id='cal_parset',
        token_task='token_cal',
        parent_dag=True,
        upload_file=args_dict['cal1_parset'],
        parset_task = 'make_parsets',
        pc_database = 'sksp2juelich',
        dag=dag)

    sandbox_cal2 = LRTSandboxOperator(task_id='sbx_cal2',
        sbx_config=args_dict['pref_cal2_cfg'],
        dag=dag)
     
    tokens_cal2 = TokenCreator( task_id='token_cal2',
        staging_task={'name':'check_calstaged','parent_dag':False},
        sbx_task={'name':'sbx_cal2','parent_dag':False},
        token_type=field_name,
        files_per_token=999,
        fields_task = {'name':'get_next_field','parent_dag':True} ,
        tok_config=args_dict['pref_cal2_cfg'],
        pc_database = 'sksp2juelich',
        dag=dag)
        
    parset_cal2 = TokenUploader( task_id='cal_parset2',
        token_task='token_cal2',
        parent_dag=True,
        upload_file=args_dict['cal2_parset'],
        parset_task = 'make_parsets',
        pc_database = 'sksp2juelich',
        dag=dag)

    sandbox_targ1 = LRTSandboxOperator(task_id='sbx_targ1',
        sbx_config=args_dict['pref_targ1_cfg'],
        trigger_rule='all_done',        # The task will start when parents are success or skipped 
        dag=dag)
        
    tokens_targ1 = TokenCreator( task_id='token_targ1',
            staging_task={'name':'stage_targ','parent_dag':False},
        sbx_task={'name':'sbx_targ1','parent_dag':False},
        srms_task={'name':'get_srmfiles','parent_dag':True},
        token_type=field_name,
        files_per_token=1,
        fields_task = {'name':'get_next_field','parent_dag':True} ,
        tok_config=args_dict['pref_targ1_cfg'],
        pc_database = 'sksp2juelich',
        dag=dag)
        
    parset_targ1 = TokenUploader( task_id='targ_parset1',
        token_task='token_targ1',
        parent_dag=True,
        upload_file=args_dict['targ1_parset'],
        parset_task = 'make_parsets',
        pc_database = 'sksp2juelich',
        dag=dag)

    branch_targ_if_staging_needed = BranchPythonOperator(
    task_id='branch_targ_if_staging_needed',
    provide_context=True,                   # Allows to access returned values from other tasks
    python_callable=stage_if_needed,
    op_args=['check_targstaged','files_staged_targ','stage_targ'],
    dag=dag) 
    
    files_staged_targ = DummyOperator(
    task_id='files_staged_targ',
    dag=dag)   
     
    
    join_targ = DummyOperator(
    task_id='join_targ',
    trigger_rule='one_success',
    dag=dag)   

    stage_targ= LOFARStagingOperator( task_id='stage_targ',
        srmfile={'name':'get_srmfiles','parent_dag':True},
        srmkey='targ_srmfile',
        dag=dag)
    
    check_targstaged = Check_staged( task_id='check_targstaged',
        srmfile={'name':'get_srmfiles','parent_dag':True},
        srmkey='targ_srmfile',
        dag=dag)
    
    check_done_files = dcacheSensor(task_id='check_done_files',
            poke_interval=1200,
            token_task = 'token_targ1',
            dag=dag)

    branch_if_cal_exists >> check_calstaged
    branch_if_cal_exists >> calib_done >>sandbox_targ1

#checking if calibrator is staged
    check_calstaged >>  branching_cal
    branching_cal >> stage >> join
    branching_cal >> files_staged >> join
 
    join >> sandbox_cal 

    sandbox_cal >> tokens_cal >> parset_cal >> sandbox_cal2 
    sandbox_cal2 >> tokens_cal2 >> parset_cal2 

    check_targstaged >> branch_targ_if_staging_needed

    branch_targ_if_staging_needed >> files_staged_targ >> join_targ
    branch_targ_if_staging_needed >> stage_targ >> join_targ

    join_targ >> sandbox_targ1 

    parset_cal2 >> sandbox_targ1 >> tokens_targ1 >> parset_targ1 >> check_done_files
    return dag



import os
from airflow import DAG
from airflow.operators.bash_operator import BashOperator
from AGLOW.airflow.operators.LTA_staging import LOFARStagingOperator
from AGLOW.airflow.operators.LRT_Sandbox import LRTSandboxOperator
from AGLOW.airflow.operators.LRT_token import TokenCreator,TokenUploader,ModifyTokenStatus
from AGLOW.airflow.operators.LRT_submit import LRTSubmit 
from AGLOW.airflow.operators.data_staged import Check_staged
from AGLOW.airflow.sensors.glite_wms_sensor import gliteSensor
from datetime import datetime, timedelta
from airflow.operators.python_operator import PythonOperator
from airflow.operators.python_operator import BranchPythonOperator
from airflow.operators.dummy_operator import DummyOperator
from AGLOW.airflow.operators.LRT_storage_to_srm import Storage_to_Srmlist 
from airflow.models import Variable

#Import helper fucntions 
from AGLOW.airflow.utils.AGLOW_utils import get_next_field
from AGLOW.airflow.utils.AGLOW_utils import count_files_uberftp 
from AGLOW.airflow.utils.AGLOW_utils import count_grid_files
from AGLOW.airflow.utils.AGLOW_utils import stage_if_needed
from AGLOW.airflow.utils.AGLOW_utils import get_next_field
from AGLOW.airflow.utils.AGLOW_utils import set_field_status_from_taskid
from AGLOW.airflow.utils.AGLOW_utils import get_srmfile_from_dir
from AGLOW.airflow.utils.AGLOW_utils import count_from_task
from AGLOW.airflow.utils.AGLOW_utils import get_field_location_from_srmlist
from AGLOW.airflow.utils.AGLOW_utils import set_field_status_from_task_return
from AGLOW.airflow.utils.AGLOW_utils import modify_parset_from_fields_task 
from AGLOW.airflow.utils.AGLOW_utils import check_folder_for_files_from_task 
#from airflow.utils.AGLOW_utils import archive_tokens_from_task
from GRID_LRT.Staging import state_all


from GRID_LRT.Staging.srmlist import srmlist
from GRID_LRT import Token
from GRID_LRT.get_picas_credentials import picas_cred
import subprocess
import  fileinput
import logging
import pdb

def test_check_staged(srm_variable, **context):
    srmfile = Variable.get(srm_variable)
    logging.info(state_all.__file__)
#    file_statuses = state_all.main(srmfile, verbose=False)
    return {'staged':False,'srmfile':str(srmfile)}

#TODO: This fails!

def archive_tokens_from_task(token_task, delete=False, **context):
    """ Determines whic tokens to archive and saves them. delete if necessary
    """
    task_dict = context['ti'].xcom_pull(token_task)
    t_type = task_dict['token_type']
    archive_location = task_dict['output_dir']
    archive_all_tokens(t_type, archive_location, delete=delete)


def archive_all_tokens(token_type, archive_location, delete=False):
    pc = picas_cred()
    th = Token.Token_Handler(t_type=token_type, uname=pc.user, pwd=pc.password, dbn=pc.database)
    token_archive = th.archive_tokens(delete_on_save=delete, compress=True)
    logging.info("Archived tokens from " + token_type + " and made an archive: " + token_archive)
    logging.info(token_archive + " size is " + str(os.stat(token_archive).st_size))
    subprocess.call(['globus-url-copy '+token_archive+" "+archive_location+"/"+token_archive.split('/')[-1]],shell=True)
    logging.info("Resulting archive is at "+archive_location+"/"+token_archive.split('/')[-1])

    

def calibrator_subdag(parent_dag_name, subdagname, dag_args, args_dict=None):
    field_name = 'fields_'
    #Resetting variables (They are set by get_srmfiles)
    
    dag = DAG(dag_id=parent_dag_name+'.'+subdagname, default_args=dag_args, schedule_interval='@once' , catchup=False)
    if not args_dict:
        args_dict = { 
            "cal1_parset":"/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/parsets/Pre-Facet-Calibrator-1.parset", 
            "cal2_parset":"/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/parsets/Pre-Facet-Calibrator-2.parset", 
            'pref_cal1_cfg':'/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/config/steps/pref_cal1.cfg',
            'pref_cal2_cfg':'/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/data/config/steps/pref_cal2.cfg'
            }

    #Create a sandbox for the job
    sandbox_cal = LRTSandboxOperator(task_id='sbx',
            sbx_config=args_dict['pref_cal1_cfg'],
            dag=dag)
    
    #Create the tokens and populate the srm.txt 
    tokens_cal = TokenCreator(task_id='token_cal',
            sbx_task={'name':'sbx', 'parent_dag':False},
            staging_task ={'name':'check_calstaged', 'parent_dag':True},
            token_type=field_name,
            tok_config =args_dict['pref_cal1_cfg'],
            files_per_token=1,
            dag=dag)
    
    #Upload the parset to all the tokens
    parset_cal = TokenUploader(task_id='cal_parset', 
            token_task='token_cal',
            upload_file=args_dict['cal1_parset'],
            parset_task = 'make_parsets',
            parent_dag = True,
            dag=dag)
    
    
    #Submit tokens to the GRID
    submit_cal = LRTSubmit(task_id='submit',
            token_task='token_cal',
            parameter_step=1,
            NCPU=2,
            dag=dag)
    
    #Wait for all jobs to finish
    wait_for_run_cal = gliteSensor(task_id='running',
            submit_task='submit',
            success_threshold=0.95,
            poke_interval=120,
            dag=dag)
        
    cal1_files_done = PythonOperator(
            task_id = 'cal1_files_done',
            python_callable = check_folder_for_files_from_task,
            op_args = ['token_cal','output_dir', 230],
            dag=dag
            )
    #####################################
    #######Calibrator 2 block
    #####################################
    
    sandbox_cal2 = LRTSandboxOperator(task_id='sbx_cal2', 
            sbx_config=args_dict['pref_cal2_cfg'],
            dag=dag)
    
    tokens_cal2 = TokenCreator( task_id='token_cal2',
            staging_task={'name':'check_calstaged','parent_dag':True},
            sbx_task={'name':'sbx_cal2','parent_dag':False},
            token_type=field_name,
            files_per_token=999,
            tok_config=args_dict['pref_cal2_cfg'],
            dag=dag)
    
    parset_cal2 = TokenUploader( task_id='cal_parset2',
            token_task='token_cal2',
            upload_file=args_dict['cal2_parset'],
            parset_task = 'make_parsets',
            parent_dag = True,
            dag=dag)
    
    submit_cal2 = LRTSubmit( task_id='submit_cal2',
            token_task='token_cal2',
            parameter_step=1, 
            NCPU=5,
            dag=dag)
    
    wait_for_run_cal2 = gliteSensor( task_id='running_cal2',
            submit_task='submit_cal2',
            success_threshold=0.9,
            poke_interval=120,
            dag=dag)
    
    archive_cal2 = PythonOperator(
            task_id = 'archive_cal2',
            provide_context=True, 
            python_callable=archive_tokens_from_task,
            op_args=['token_cal2'],
            dag = dag)
     

    ## Setting up the dependency graph of the workflow
    
    
    #Branch if calibrator already processed

    sandbox_cal >> tokens_cal >> parset_cal >>  submit_cal >> wait_for_run_cal
    wait_for_run_cal >> cal1_files_done >> sandbox_cal2 >> tokens_cal2 >> parset_cal2 >> submit_cal2 >> wait_for_run_cal2
    
    wait_for_run_cal2 >> archive_cal2 
    return dag
